-- 
-- Copyright (C) 2006 Johannes Hausensteiner (johannes.hausensteiner@pcl.at)
-- 
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
--
-- 
-- Filename: cpu.vhd
--
-- Function: the DIY calculator CPU
-- 
--
-- Changelog
--
--  1.0    4.Aug.2006   JH   new
--  1.1   23.Aug.2006   JH   new design
--  1.2   30.Aug.2006   JH   code restructure, JIR instruction, nWAIT
--  1.3    6.Sep.2006   JH   VER instruction
--


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity CPU is
  port (
    CLK    : in std_logic;
    nRST   : in std_logic;
    d_in   : in std_logic_vector (7 downto 0);
    d_out  : out std_logic_vector (7 downto 0);
    addr   : out std_logic_vector (15 downto 0);
    nIRQ   : in std_logic;
    nIACK  : out std_logic;
    nWAIT  : in std_logic;
    M1     : out std_logic;
    nRD    : out std_logic;
    nWR    : out std_logic;
    ddrive : out std_logic
  );
end CPU;

architecture behave of CPU is
  -- define the states we use
	type state_t is (
	  T0A, T0B, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14
	);
	constant CPU_REVISION : std_logic_vector (7 downto 0) := x"03";

	-- the CPU registers
	signal acc : std_logic_vector (7 downto 0);  -- accumulator
	signal sr  : std_logic_vector (4 downto 0);  -- status register
	signal b   : std_logic_vector (7 downto 0);  -- 2nd input to ALU
  signal ir  : std_logic_vector (7 downto 0);  -- instruction register
	signal x   : std_logic_vector (15 downto 0); -- X register
	signal iv  : std_logic_vector (15 downto 0); -- interrupt vector reg.
	signal sp  : std_logic_vector (15 downto 0); -- stack pointer
	signal tpca: std_logic_vector (15 downto 0); -- temp. pgm counter A
	signal tpcb: std_logic_vector (15 downto 0); -- temp. pgm counter B
	signal pc  : std_logic_vector (15 downto 0); -- program counter
	
	-- bits of the status register
	alias sr_c : std_logic is sr(0);  -- carry flag
	alias sr_z : std_logic is sr(1);  -- zero flag
	alias sr_n : std_logic is sr(2);  -- negative flag
	alias sr_o : std_logic is sr(3);  -- overflow flag
	alias sr_i : std_logic is sr(4);  -- interrupt mask flag

	signal irq  : std_logic;  -- interrupt latch
	signal iack : std_logic;  -- interrupt acknowledge

	-- I/O registers
	signal din : std_logic_vector (7 downto 0);  -- data input buffer
	signal dout: std_logic_vector (7 downto 0);  -- data output buffer
	signal aout: std_logic_vector (15 downto 0); -- address output buffer

	-- state machine
	signal state : state_t;       -- current state
	signal next_state: state_t;   -- next state

	-- the CPU instructions
	-- addressing mode suffix:
	--   i ... immediate
	--   a ... absolute
	--   x ... indexed
	--   n ... indirect
	--   xi .. pre-indexed indirect
	--   ix .. indirect post-indexed

  -- control
	constant NOP    : std_logic_vector := x"00";
	constant HALT   : std_logic_vector := x"01";
	constant SETIM  : std_logic_vector := x"08";
	constant CLRIM  : std_logic_vector := x"09";

  -- arithmetic
	constant ADDi   : std_logic_vector := x"10";
	constant ADDa   : std_logic_vector := x"11";
	constant ADDx   : std_logic_vector := x"12";

	constant ADDCi  : std_logic_vector := x"18";
	constant ADDCa  : std_logic_vector := x"19";
	constant ADDCx  : std_logic_vector := x"1A";

	constant SUBi   : std_logic_vector := x"20";
	constant SUBa   : std_logic_vector := x"21";
	constant SUBx   : std_logic_vector := x"22";

	constant SUBCi  : std_logic_vector := x"28";
	constant SUBCa  : std_logic_vector := x"29";
	constant SUBCx  : std_logic_vector := x"2A";

  -- logical
	constant ANDi   : std_logic_vector := x"30";
	constant ANDa   : std_logic_vector := x"31";
	constant ANDx   : std_logic_vector := x"32";

	constant ORi    : std_logic_vector := x"38";
	constant ORa    : std_logic_vector := x"39";
	constant ORx    : std_logic_vector := x"3A";

	constant XORi   : std_logic_vector := x"40";
	constant XORa   : std_logic_vector := x"41";
	constant XORx   : std_logic_vector := x"42";

  -- comparison
	constant CMPAi  : std_logic_vector := x"60";
	constant CMPAa  : std_logic_vector := x"61";
	constant CMPAx  : std_logic_vector := x"62";

  -- shift/rotate
	constant SHL    : std_logic_vector := x"70";
	constant SHR    : std_logic_vector := x"71";
	constant ROLC   : std_logic_vector := x"78";
	constant RORC   : std_logic_vector := x"79";

  -- increment/decrement
	constant INCA   : std_logic_vector := x"80";
	constant DECA   : std_logic_vector := x"81";
	constant INCX   : std_logic_vector := x"82";
	constant DECX   : std_logic_vector := x"83";

  -- load/store   
	constant LDAi   : std_logic_vector := x"90";
	constant LDAa   : std_logic_vector := x"91";
	constant LDAx   : std_logic_vector := x"92";
	constant LDAn   : std_logic_vector := x"93";
	constant LDAxi  : std_logic_vector := x"94";
	constant LDAix  : std_logic_vector := x"95";

	constant STAa   : std_logic_vector := x"99";
	constant STAx   : std_logic_vector := x"9A";
	constant STAn   : std_logic_vector := x"9B";
	constant STAxi  : std_logic_vector := x"9C";
	constant STAix  : std_logic_vector := x"9D";

	constant BLDXi  : std_logic_vector := x"A0";
	constant BLDXa  : std_logic_vector := x"A1";

	constant BSTXa  : std_logic_vector := x"A9";

	constant BLDSPi : std_logic_vector := x"50";
	constant BLDSPa : std_logic_vector := x"51";

	constant BSTSPa : std_logic_vector := x"59";

	constant BLDIVi : std_logic_vector := x"F0";
	constant BLDIVa : std_logic_vector := x"F1";

  -- push/pop
	constant POPA   : std_logic_vector := x"B0";
	constant PUSHA  : std_logic_vector := x"B2";
	constant POPSR  : std_logic_vector := x"B1";
	constant PUSHSR : std_logic_vector := x"B3";

  -- unconditional jump
	constant JMPa   : std_logic_vector := x"C1";
	constant JMPx   : std_logic_vector := x"C2";
	constant JMPn   : std_logic_vector := x"C3";
	constant JMPxi  : std_logic_vector := x"C4";
	constant JMPix  : std_logic_vector := x"C5";

	constant JSRa   : std_logic_vector := x"C9";
	constant JSRx   : std_logic_vector := x"CA";
	constant JSRn   : std_logic_vector := x"CB";
	constant JSRxi  : std_logic_vector := x"CC";
	constant JSRix  : std_logic_vector := x"CD";

  -- conditional jump, absolute addressing only
	constant JC     : std_logic_vector := x"E1";
	constant JNC    : std_logic_vector := x"E6";
	constant JZ     : std_logic_vector := x"D1";
	constant JNZ    : std_logic_vector := x"D6";
	constant JN     : std_logic_vector := x"D9";
	constant JNN    : std_logic_vector := x"DE";
	constant JO     : std_logic_vector := x"E9";
	constant JNO    : std_logic_vector := x"EE";

  -- return
	constant RTS    : std_logic_vector := x"CF";
	constant RTI    : std_logic_vector := x"C7";

  -- interrupt
  constant JIR    : std_logic_vector := x"FF";

  -- read CPU version
	constant VER    : std_logic_vector := x"98";

  -- accumulator control
  constant ACC_NOP  : integer := 0;
  constant ACC_LOAD : integer := 1;
  constant ACC_INC  : integer := 2;
  constant ACC_DEC  : integer := 3;
  constant ACC_SHL  : integer := 4;
  constant ACC_SHR  : integer := 5;
  constant ACC_ROLC : integer := 6;
  constant ACC_RORC : integer := 7;
  constant ACC_ADD  : integer := 8;
  constant ACC_ADDC : integer := 9;
  constant ACC_SUB  : integer := 10;
  constant ACC_SUBC : integer := 11;
  constant ACC_AND  : integer := 12;
  constant ACC_OR   : integer := 13;
  constant ACC_XOR  : integer := 14;
  constant ACC_VER  : integer := 15;
  signal acc_ctrl : integer;

  -- carry flag control
  constant SR_C_NOP : integer := 0;
  constant SR_C_SET : integer := 1;
  constant SR_C_CLR : integer := 2;
  signal sr_c_ctrl : integer;

  -- zero flag control
  constant SR_Z_NOP : integer := 0;
  constant SR_Z_SET : integer := 1;
  constant SR_Z_CLR : integer := 2;
  signal sr_z_ctrl : integer;

  -- negative flag control
  constant SR_N_NOP : integer := 0;
  constant SR_N_SET : integer := 1;
  constant SR_N_CLR : integer := 2;
  signal sr_n_ctrl : integer;

  -- overflow flag control
  constant SR_O_NOP : integer := 0;
  constant SR_O_SET : integer := 1;
  constant SR_O_CLR : integer := 2;
  signal sr_o_ctrl : integer;

  -- interrupt mask flag control
  constant SR_I_NOP : integer := 0;
  constant SR_I_SET : integer := 1;
  constant SR_I_CLR : integer := 2;
  signal sr_i_ctrl : integer;

  -- interrupt latch control
  signal irq_rst : boolean;

  -- interrupt acknowledge control
  constant IACK_NOP : integer := 0;
  constant IACK_SET : integer := 1;
  constant IACK_CLR : integer := 2;
  signal iack_ctrl : integer;

  -- B register control
  constant B_NOP    : integer := 0;
  constant B_LOAD   : integer := 1;
  signal b_ctrl : integer;

  -- IR register control
  constant IR_NOP   : integer := 0;
  constant IR_LOAD  : integer := 1;
  constant IR_LDJIR : integer := 2;
  signal ir_ctrl : integer;

  -- stack pointer control
  constant SP_NOP   : integer := 0;
  constant SP_LDHI  : integer := 1;
  constant SP_LDLO  : integer := 2;
  constant SP_INC   : integer := 3;
  constant SP_DEC   : integer := 4;
  signal sp_ctrl : integer;

  -- program counter control
  constant PC_NOP    : integer := 0;
  constant PC_INC    : integer := 1;
  constant PC_LOADA  : integer := 2;
  constant PC_LOADAX : integer := 3;
  constant PC_LOADB  : integer := 4;
  constant PC_LOADBX : integer := 5;
  constant PC_LOADIV : integer := 6;
  signal pc_ctrl : integer;

  -- X register control
  constant X_NOP  : integer := 0;
  constant X_LDHI : integer := 1;
  constant X_LDLO : integer := 2;
  constant X_INC  : integer := 3;
  constant X_DEC  : integer := 4;
  signal x_ctrl : integer;

  -- IV register control
  constant IV_NOP  : integer := 0;
  constant IV_LDHI : integer := 1;
  constant IV_LDLO : integer := 2;
  signal iv_ctrl : integer;

  -- temp PC A register control
  constant TPCA_NOP  : integer := 0;
  constant TPCA_LDHI : integer := 1;
  constant TPCA_LDLO : integer := 2;
  constant TPCA_INC  : integer := 3;
  constant TPCA_ADDX : integer := 4;
  signal tpca_ctrl : integer;

  -- temp PC B register control
  constant TPCB_NOP  : integer := 0;
  constant TPCB_LDHI : integer := 1;
  constant TPCB_LDLO : integer := 2;
  signal tpcb_ctrl : integer;

  -- internal control signals
  signal wr_int, wr_int_d : std_logic;
  signal rd_int, rd_int_d : std_logic;

begin
  -- assign signals
  d_out <= dout;
  addr <= aout;
  nIACK <= not (iack);
  nWR <= not (wr_int);
  nRD <= not (rd_int);

  --------------------------------------------------------------------------
  --                     modelling the CPU registers                      --
  --------------------------------------------------------------------------
  -- the accumulator  ------------------------------------------------------
  acc_reg : process (CLK, nRST, acc_ctrl)
  begin
    if nRST = '0' then
      acc <= x"00";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case acc_ctrl is
        when ACC_LOAD =>
          acc <= din;
        when ACC_INC =>
          acc <= acc + 1;
        when ACC_DEC =>
          acc <= acc - 1;
        when ACC_SHL =>
          acc(7) <= acc(6);
          acc(6) <= acc(5);
          acc(5) <= acc(4);
          acc(4) <= acc(3);
          acc(3) <= acc(2);
          acc(2) <= acc(1);
          acc(1) <= acc(0);
          acc(0) <= '0';
        when ACC_SHR =>
          acc(0) <= acc(1);
          acc(1) <= acc(2);
          acc(2) <= acc(3);
          acc(3) <= acc(4);
          acc(4) <= acc(5);
          acc(5) <= acc(6);
          acc(6) <= acc(7);
        when ACC_ROLC =>
          acc(7) <= acc(6);
          acc(6) <= acc(5);
          acc(5) <= acc(4);
          acc(4) <= acc(3);
          acc(3) <= acc(2);
          acc(2) <= acc(1);
          acc(1) <= acc(0);
          acc(0) <= sr_c;
        when ACC_RORC =>
          acc(0) <= acc(1);
          acc(1) <= acc(2);
          acc(2) <= acc(3);
          acc(3) <= acc(4);
          acc(4) <= acc(5);
          acc(5) <= acc(6);
          acc(6) <= acc(7);
          acc(7) <= sr_c;
        when ACC_ADD =>
          acc <= acc + b;
        when ACC_ADDC =>
          if sr_c = '1' then acc <= acc + b + 1;
                        else acc <= acc + b;
          end if;
        when ACC_SUB =>
          acc <= acc + not(b) + 1;
        when ACC_SUBC =>
          if sr_c = '1' then acc <= acc + not(b) + 1;
                        else acc <= acc + not(b);
          end if;
        when ACC_AND =>
          acc <= acc and b;
        when ACC_OR =>
          acc <= acc or b;
        when ACC_XOR =>
          acc <= acc xor b;
        when ACC_VER =>
          acc <= CPU_REVISION;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the carry flag --------------------------------------------------------
  sr_c_bit : process (CLK, nRST, sr_c_ctrl)
  begin
    if nRST = '0' then
      sr_c <= '0';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sr_c_ctrl is
        when SR_C_SET => sr_c <= '1';
        when SR_C_CLR => sr_c <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the zero flag ---------------------------------------------------------
  sr_z_bit : process (CLK, nRST, sr_z_ctrl)
  begin
    if nRST = '0' then
      sr_z <= '1';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sr_z_ctrl is
        when SR_Z_SET => sr_z <= '1';
        when SR_Z_CLR => sr_z <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the negative flag -----------------------------------------------------
  sr_n_bit : process (CLK, nRST, sr_n_ctrl)
  begin
    if nRST = '0' then
      sr_n <= '0';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sr_n_ctrl is
        when SR_N_SET => sr_n <= '1';
        when SR_N_CLR => sr_n <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the overflow flag -----------------------------------------------------
  sr_o_bit : process (CLK, nRST, sr_o_ctrl)
  begin
    if nRST = '0' then
      sr_o <= '0';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sr_o_ctrl is
        when SR_O_SET => sr_o <= '1';
        when SR_O_CLR => sr_o <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the interrupt mask flag -----------------------------------------------
  sr_i_bit : process (CLK, nRST, sr_i_ctrl)
  begin
    if nRST = '0' then
      sr_i <= '0';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sr_i_ctrl is
        when SR_I_SET => sr_i <= '1';
        when SR_I_CLR => sr_i <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the interrupt latch ---------------------------------------------------
  irq_latch : process (CLK, nRST, nIRQ, irq_rst)
  begin
    if nRST = '0' then
      irq <= '0';
    elsif rising_edge (CLK) then
      if nIRQ = '0' then irq <= '1';
      elsif irq_rst = true then irq <= '0';
      end if;
    end if;
  end process;

  -- the interrupt acknowledge flag ----------------------------------------
  iack_bit : process (CLK, nRST, iack_ctrl)
  begin
    if nRST = '0' then
      iack <= '0';
    elsif rising_edge (CLK) and nWAIT = '1' then
      case iack_ctrl is
        when IACK_SET => iack <= '1';
        when IACK_CLR => iack <= '0';
        when others => NULL;
      end case;
    end if;
  end process;

  -- the instruction register ----------------------------------------------
  ir_reg : process (CLK, nRST, ir_ctrl)
  begin
    if nRST = '0' then
      ir <= x"00";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case ir_ctrl is
        when IR_LOAD => ir <= din;
        when IR_LDJIR => ir <= JIR;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the B register --------------------------------------------------------
  b_reg : process (CLK, nRST, b_ctrl)
  begin
    if nRST = '0' then
      b <= x"00";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case b_ctrl is
        when B_LOAD => b <= din;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the program counter ---------------------------------------------------
  pc_reg : process (CLK, nRST, pc_ctrl)
  begin
    if nRST = '0' then
      pc <= x"0000";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case pc_ctrl is
        when PC_INC =>
          pc <= pc + 1;
        when PC_LOADA =>
          pc <= tpca;
        when PC_LOADAX =>
          pc <= tpca + x;
        when PC_LOADB =>
          pc <= tpcb;
        when PC_LOADBX =>
          pc <= tpcb + x;
        when PC_LOADIV =>
          pc <= iv;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the stack pointer -----------------------------------------------------
  sp_reg : process (CLK, nRST, sp_ctrl)
  begin
    if nRST = '0' then
      sp <= x"0000";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case sp_ctrl is
        when SP_INC => sp <= sp + 1;
        when SP_DEC => sp <= sp - 1;
        when SP_LDHI => sp(15 downto 8) <= din;
        when SP_LDLO => sp (7 downto 0) <= din;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the X register --------------------------------------------------------
  x_reg : process (CLK, nRST, x_ctrl)
  begin
    if nRST = '0' then
      x <= x"5555";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case x_ctrl is
        when X_INC => x <= x + 1;
        when X_DEC => x <= x - 1;
        when X_LDHI => x(15 downto 8) <= din;
        when X_LDLO => x (7 downto 0) <= din;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the IV register -------------------------------------------------------
  iv_reg : process (CLK, nRST, iv_ctrl)
  begin
    if nRST = '0' then
      iv <= x"0000";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case iv_ctrl is
        when IV_LDHI => iv(15 downto 8) <= din;
        when IV_LDLO => iv (7 downto 0) <= din;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the temp PC A ---------------------------------------------------------
  tpca_reg : process (CLK, nRST, tpca_ctrl)
  begin
    if nRST = '0' then
      tpca <= x"0000";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case tpca_ctrl is
        when TPCA_LDHI => tpca(15 downto 8) <= din;
        when TPCA_LDLO => tpca (7 downto 0) <= din;
        when TPCA_INC  => tpca <= tpca + 1;
        when TPCA_ADDX => tpca <= tpca + x;
        when others => NULL;
      end case;
    end if;
  end process;

  -- the temp PC B ---------------------------------------------------------
  tpcb_reg : process (CLK, nRST, tpcb_ctrl)
  begin
    if nRST = '0' then
      tpcb <= x"0000";
    elsif rising_edge (CLK) and nWAIT = '1' then
      case tpcb_ctrl is
        when TPCB_LDHI => tpcb(15 downto 8) <= din;
        when TPCB_LDLO => tpcb (7 downto 0) <= din;
        when others => NULL;
      end case;
    end if;
  end process;

  --------------------------------------------------------------------------
  --                      modelling control signals                       --
  --------------------------------------------------------------------------
  -- delay the read signal -------------------------------------------------
  rd_delay : process (CLK, nRST, rd_int_d)
  begin
    if nRST = '0' then
      rd_int <= '0';
    elsif falling_edge (CLK) then
      rd_int <= rd_int_d;
    end if;
  end process;

  -- delay the write signal ------------------------------------------------
  wr_delay : process (CLK, nRST, wr_int_d)
  begin
    if nRST = '0' then
      wr_int <= '0';
    elsif falling_edge (CLK) then
      wr_int <= wr_int_d;
    end if;
  end process;

  -- latch input data ------------------------------------------------------
  din_reg : process (rd_int, nRST)
  begin
    if nRST = '0' then
      din <= x"00";
    elsif falling_edge (rd_int) then
      din <= d_in;
    end if;
  end process;

  --------------------------------------------------------------------------
  --                   modelling the CPU state machine                    --
  --------------------------------------------------------------------------
  -- advance the state machine ---------------------------------------------
  process (CLK, nRST, nWAIT)
  begin
    if nRST = '0' then
      state <= T0A;
    elsif rising_edge (CLK) then
      if nWAIT = '1' then state <= next_state;
      end if;
    end if;
  end process;

  -- transition table ------------------------------------------------------
  process (state, ir, sr, irq)
  begin       
    case state is
      ----------------------------------------------------------------------
      -- CPU state T0
      ----------------------------------------------------------------------
      when T0A => next_state <= T0B;
      when T0B => next_state <= T1;

      ----------------------------------------------------------------------
      -- CPU state T1
      ----------------------------------------------------------------------
      when T1 => next_state <= T2;

      ----------------------------------------------------------------------
      -- CPU state T2
      ----------------------------------------------------------------------
      when T2 => next_state <= T3;

      ----------------------------------------------------------------------
      -- CPU state T3
      ----------------------------------------------------------------------
      when T3 =>
        case ir is
          when INCA|DECA|INCX|DECX|SHL|SHR|ROLC|RORC|SETIM|CLRIM|NOP|VER =>
            next_state <= T1;
          when ADDi | ADDCi | SUBi | SUBCi | ANDi | ORi | XORi | CMPAi =>
            next_state <= T8;
          when HALT =>
            if sr_i = '1' and irq = '1' then next_state <= T1;
                                        else next_state <= T3;
            end if;
          when others => next_state <= T4;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T4
      ----------------------------------------------------------------------
      when T4 =>
        case ir is
          when POPA | POPSR | PUSHA | PUSHSR | LDAi =>
            next_state <= T1;
          when JC     =>
            if sr_c = '1' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JNC    =>
            if sr_c = '0' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JZ     =>
            if sr_z = '1' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JNZ    =>
            if sr_z = '0' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JN     =>
            if sr_n = '1' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JNN    =>
            if sr_n = '0' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JO     =>
            if sr_o = '1' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when JNO    =>
            if sr_o = '0' then next_state <= T5;
                          else next_state <= T1;
            end if;
          when others => next_state <= T5;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T5
      ----------------------------------------------------------------------
      when T5 => next_state <= T6;

      ----------------------------------------------------------------------
      -- CPU state T6
      ----------------------------------------------------------------------
      when T6 =>
        case ir is
          when BLDIVi|BLDSPi|BLDXi|JMPa|JC|JNC|JZ|JNZ|JN|JNN|JO|JNO|JMPx|RTS=>
            next_state <= T1;
          when others => next_state <= T7;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T7
      ----------------------------------------------------------------------
      when T7 => next_state <= T8;

      ----------------------------------------------------------------------
      -- CPU state T8
      ----------------------------------------------------------------------
      when T8 =>
        case ir is
          when ADDi|ADDCi|SUBi|SUBCi|ANDi|ORi|XORi|CMPAi|LDAi|
               ADDa|ADDCa|SUBa|SUBCa|ANDa|ORa|XORa|CMPAa|LDAa|STAa|RTI|
               ADDx|ADDCx|SUBx|SUBCx|ANDx|ORx|XORx|CMPAx|LDAx|STAx|JIR =>
            next_state <= T1;
          when others => next_state <= T9;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T9
      ----------------------------------------------------------------------
      when T9 => next_state <= T10;

      ----------------------------------------------------------------------
      -- CPU state T10
      ----------------------------------------------------------------------
      when T10 =>
        case ir is
          when BLDIVa|BLDSPa|BLDXa|BSTSPa|BSTXa|JMPn|JMPxi|JMPix|JSRa|JSRx =>
            next_state <= T1;
          when others => next_state <= T11;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T11
      ----------------------------------------------------------------------
      when T11 => next_state <= T12;

      ----------------------------------------------------------------------
      -- CPU state T12
      ----------------------------------------------------------------------
      when T12 =>
        case ir is
          when LDAn | LDAxi | LDAix | STAn | STAxi | STAix =>
            next_state <= T1;
          when others => next_state <= T13;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T13
      ----------------------------------------------------------------------
      when T13 => next_state <= T14;

      ----------------------------------------------------------------------
      -- CPU state T14
      ----------------------------------------------------------------------
      when T14 =>
       case ir is
         when JSRn | JSRxi | JSRix =>
           next_state <= T1;
         when others => next_state <= T1;
       end case;

      when others => next_state <= T1;
    end case;
  end process;  -- end transition table

  -- output table ----------------------------------------------------------
  process (state)
  begin
    -- default values
--    nRD <= '1';
--    nWR <= '1';
    M1 <= '1';
    ddrive <= '0';
    dout <= x"FF";
    aout <= pc;
    b_ctrl <= B_NOP;
    ir_ctrl <= IR_NOP;
    acc_ctrl <= ACC_NOP;
    sp_ctrl <= SP_NOP;
    pc_ctrl <= PC_NOP;
    x_ctrl <= X_NOP;
    iv_ctrl <= IV_NOP;
    tpca_ctrl <= TPCA_NOP;
    tpcb_ctrl <= TPCB_NOP;
    sr_c_ctrl <= SR_C_NOP;
    sr_z_ctrl <= SR_Z_NOP;
    sr_n_ctrl <= SR_N_NOP;
    sr_o_ctrl <= SR_O_NOP;
    sr_i_ctrl <= SR_I_NOP;
    iack_ctrl <= IACK_NOP;
    irq_rst <= false;
    wr_int_d <= '0';
    rd_int_d <= '0';

    case state is
      ----------------------------------------------------------------------
      -- CPU state T0
      ----------------------------------------------------------------------
      when T0A =>  -- reset 1
        null;
      when T0B =>  -- reset 2
        rd_int_d <= '1';

      ----------------------------------------------------------------------
      -- CPU state T1
      ----------------------------------------------------------------------
      when T1 =>   -- opcode fetch 1
        M1 <= '0';
        if sr_i = '1' and irq = '1' then
          -- HW interrupt detected
          ir_ctrl <= IR_LDJIR;
        else
--          nRD <= '0';
          ir_ctrl <= IR_LOAD;
          pc_ctrl <= PC_INC;
        end if;

      ----------------------------------------------------------------------
      -- CPU state T2
      ----------------------------------------------------------------------
      when T2 =>   -- opcode fetch 2
        case ir is
          when JIR =>
            ddrive <= '1';
            wr_int_d <= '1';
            sr_i_ctrl <= SR_I_CLR;
            aout <= sp;
          when PUSHA | PUSHSR =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= sp;
          when POPA | POPSR | RTS | RTI =>
            rd_int_d <= '1';
            aout <= sp + 1;
          when JC =>
            if sr_c = '1' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JNC =>
            if sr_c = '0' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JZ =>
            if sr_z = '1' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JNZ =>
            if sr_z = '0' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JN =>
            if sr_n = '1' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JNN =>
            if sr_n = '0' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JO =>
            if sr_o = '1' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when JNO =>
            if sr_o = '0' then rd_int_d <= '1';
                          else pc_ctrl <= PC_INC;
            end if;
          when ADDi | ADDCi | SUBi | SUBCi | ANDi | ORi | XORi | CMPAi |LDAi|
               ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa |
               ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx |
               BLDIVi|BLDSPi|BLDXi|BLDIVa|BLDSPa|BLDXa|BSTSPa|BSTXa|
               LDAa | LDAx | LDAn | LDAxi | LDAix |
               STAa | STAx | STAn | STAxi | STAix |
               JMPa | JMPx | JMPn | JMPxi | JMPix |
               JSRa | JSRx | JSRn | JSRxi | JSRix =>
            rd_int_d <= '1';

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T3
      ----------------------------------------------------------------------
      when T3 =>   -- decode and start execution
        case ir is
          when JIR =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc (7 downto 0);
            sp_ctrl <= SP_DEC;
            irq_rst <= true;
            iack_ctrl <= IACK_SET;
            aout <= sp;
          when NOP =>
            rd_int_d <= '1';
          when VER =>
            acc_ctrl <= ACC_VER;
            rd_int_d <= '1';
          when INCA =>
            acc_ctrl <= ACC_INC;
            -- update zero flag
            if acc = x"ff" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc >= x"7f" and acc /= x"ff" then
                                 sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when DECA =>
            acc_ctrl <= ACC_DEC;
            -- update zero flag
            if acc = x"01" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc >= x"81" or acc = x"00" then
                                 sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when INCX =>
            x_ctrl <= X_INC;
            -- update zero flag
            if x = x"ffff" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            rd_int_d <= '1';
          when DECX =>
            x_ctrl <= X_DEC;
            -- update zero flag
            if x = x"0001" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            rd_int_d <= '1';
          when ROLC =>
            acc_ctrl <= ACC_ROLC;
            -- update carry flag
            if acc(7) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if acc = x"80" and sr_c = '0' then
                                sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc(6) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when RORC =>
            acc_ctrl <= ACC_RORC;
            -- update carry flag
            if acc(0) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if acc = x"01" and sr_c = '0' then
                                sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if sr_c = '1' then sr_n_ctrl <= SR_N_SET;
                          else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when SHL =>
            acc_ctrl <= ACC_SHL;
            -- update carry flag
            if acc(7) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if acc = x"80" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc(6) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when SHR =>
            acc_ctrl <= ACC_SHR;
            -- update carry flag
            if acc(0) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if acc = x"01" then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when SETIM =>
            sr_i_ctrl <= SR_I_SET;
            irq_rst <= true;
            rd_int_d <= '1';
          when CLRIM =>
            sr_i_ctrl <= SR_I_CLR;
            irq_rst <= true;
            rd_int_d <= '1';
          when POPA =>
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            aout <= sp + 1;
            sp_ctrl <= SP_INC;
          when POPSR =>
--            nRD <= '0';
            aout <= sp + 1;
            sp_ctrl <= SP_INC;
          when PUSHA =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= acc;
            sp_ctrl <= SP_DEC;
            aout <= sp;
          when PUSHSR =>
--            nWR <= '0';
            ddrive <= '1';
            dout(4 downto 0) <= sr;
            dout(7 downto 5) <= "000";
            sp_ctrl <= SP_DEC;
            aout <= sp;
          when RTS =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDHI;
            sp_ctrl <= SP_INC;
            aout <= sp + 1;
          when RTI =>
--            nRD <= '0';
            sp_ctrl <= SP_INC;
            aout <= sp + 1;

          when ADDi | ADDCi | SUBi | SUBCi | ANDi | ORi | XORi | CMPAi =>
--            nRD <= '0';
            b_ctrl <= B_LOAD;
            pc_ctrl <= PC_INC;
          when LDAi =>  -- update flags next T state
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            pc_ctrl <= PC_INC;

          when BLDIVi =>
--            nRD <= '0';
            iv_ctrl <= IV_LDHI;
            pc_ctrl <= PC_INC;
          when BLDSPi =>
--            nRD <= '0';
            sp_ctrl <= SP_LDHI;
            pc_ctrl <= PC_INC;
          when BLDXi =>
--            nRD <= '0';
            x_ctrl <= X_LDHI;
            pc_ctrl <= PC_INC;

          when JC =>
            if sr_c = '1' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JNC =>
            if sr_c = '0' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JZ =>
            if sr_z = '1' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JNZ =>
            if sr_z = '0' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JN =>
            if sr_n = '1' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JNN =>
            if sr_n = '0' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JO =>
            if sr_o = '1' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;
          when JNO =>
            if sr_o = '0' then tpca_ctrl <= TPCA_LDHI;
            end if;
            pc_ctrl <= PC_INC;

          when ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa |
               ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx |
               BLDIVa | BLDSPa | BLDXa | BSTSPa | BSTXa |
               LDAa | LDAx | LDAn | LDAxi | LDAix |
               STAa | STAx | STAn | STAxi | STAix |
               JMPa | JMPx | JMPn | JMPxi | JMPix |
               JSRa | JSRx | JSRn | JSRxi | JSRix =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDHI;
            pc_ctrl <= PC_INC;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T4
      ----------------------------------------------------------------------
      when T4 =>
        case ir is
          when JIR =>
            ddrive <= '1';
            wr_int_d <= '1';
            irq_rst <= true;
            aout <= sp;
          when POPSR =>
            if din(0) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            if din(1) = '1' then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            if din(2) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            if din(3) = '1' then sr_o_ctrl <= SR_O_SET;
                            else sr_o_ctrl <= SR_O_CLR;
            end if;
            if din(4) = '1' then sr_i_ctrl <= SR_I_SET;
                            else sr_i_ctrl <= SR_I_CLR;
            end if;
            rd_int_d <= '1';
          when POPA | PUSHA | PUSHSR =>
            rd_int_d <= '1';
          when RTS =>
            rd_int_d <= '1';
            sp_ctrl <= SP_INC;
            aout <= sp + 1;
          when RTI =>
            if din(0) = '1' then sr_c_ctrl <= SR_C_SET;
                            else sr_c_ctrl <= SR_C_CLR;
            end if;
            if din(1) = '1' then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            if din(2) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            if din(3) = '1' then sr_o_ctrl <= SR_O_SET;
                            else sr_o_ctrl <= SR_O_CLR;
            end if;
            sr_i_ctrl <= SR_I_SET;
            irq_rst <= true;
            rd_int_d <= '1';
            sp_ctrl <= SP_INC;
            aout <= sp + 1;

          when LDAi =>
            -- update zero flag
            if din = x"00"  then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if din(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';

          when ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa |
               ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx |
               BLDIVi|BLDSPi|BLDXi|BLDIVa|BLDSPa|BLDXa|BSTSPa|BSTXa|
               JC | JNC | JZ | JNZ | JN | JNN | JO | JNO |
               LDAa | LDAx | LDAn | LDAxi | LDAix |
               STAa | STAx | STAn | STAxi | STAix |
               JMPa | JMPx | JMPn | JMPxi | JMPix |
               JSRa | JSRx | JSRn | JSRxi | JSRix =>
            rd_int_d <= '1';
                            
          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T5
      ----------------------------------------------------------------------
      when T5 =>
        case ir is
          when JIR =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc (15 downto 8);
            sp_ctrl <= SP_DEC;
            irq_rst <= true;
            aout <= sp;
          when RTS =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDLO;
            aout <= sp;
          when RTI =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDHI;
            irq_rst <= true;
            sp_ctrl <= SP_INC;
            aout <= sp;

          when BLDIVi =>
--            nRD <= '0';
            iv_ctrl <= IV_LDLO;
            pc_ctrl <= PC_INC;
          when BLDSPi =>
--            nRD <= '0';
            sp_ctrl <= SP_LDLO;
            pc_ctrl <= PC_INC;
          when BLDXi =>
--            nRD <= '0';
            x_ctrl <= X_LDLO;
            pc_ctrl <= PC_INC;

          when ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa |
               ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx |
               BLDIVa | BLDSPa | BLDXa | BSTSPa | BSTXa |
               JC | JNC | JZ | JNZ | JN | JNN | JO | JNO |
               LDAa | LDAx | LDAn | LDAxi | LDAix |
               STAa | STAx | STAn | STAxi | STAix |
               JMPa | JMPx | JMPn | JMPxi | JMPix |
               JSRa | JSRx | JSRn | JSRxi | JSRix =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDLO;
            pc_ctrl <= PC_INC;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T6
      ----------------------------------------------------------------------
      when T6 =>
        case ir is
          when JIR =>
            ddrive <= '1';
            wr_int_d <= '1';
            irq_rst <= true;
            aout <= sp;
          when RTS =>
            pc_ctrl <= PC_LOADA;
            aout <= tpca;
            rd_int_d <= '1';
          when RTI =>
            irq_rst <= true;
            rd_int_d <= '1';
            aout <= sp;

          when BLDIVi | BLDSPi | BLDXi =>
            rd_int_d <= '1';
          when BLDIVa | BLDSPa | BLDXa =>
            rd_int_d <= '1';
            aout <= tpca;
          when BSTSPa | BSTXa =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpca;

          when ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa |
               LDAa | LDAn | STAn | LDAix | STAix |
               JMPn | JSRn | JMPix | JSRix =>
            rd_int_d <= '1';
            aout <= tpca;
          when ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx|LDAx=>
            rd_int_d <= '1';
            aout <= tpca + x;
          when LDAxi | STAxi | JMPxi | JSRxi =>
            rd_int_d <= '1';
            tpca_ctrl <= TPCA_ADDX;
            aout <= tpca + x;
          when STAa =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpca;
          when STAx =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpca + x;

          when JMPa =>
            pc_ctrl <= PC_LOADA;
            aout <= tpca;
            rd_int_d <= '1';
          when JMPx =>
            pc_ctrl <= PC_LOADAX;
            aout <= tpca + x;
            rd_int_d <= '1';
          when JC | JNC | JZ | JNZ | JN | JNN | JO | JNO =>
            pc_ctrl <= PC_LOADA;
            aout <= tpca;
            rd_int_d <= '1';

          when JSRa | JSRx =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T7
      ----------------------------------------------------------------------
      when T7 =>
        case ir is
          when JIR =>
--            nWR <= '0';
            ddrive <= '1';
            dout(3 downto 0) <= sr(3 downto 0);
            dout(7 downto 4) <= "0001";
            irq_rst <= true;
            aout <= sp;
          when RTI =>
--            nRD <= '0';
            tpca_ctrl <= TPCA_LDLO;
            irq_rst <= true;
            aout <= sp;
          when ADDa | ADDCa | SUBa | SUBCa | ANDa | ORa | XORa | CMPAa =>
--            nRD <= '0';
            b_ctrl <= B_LOAD;
            aout <= tpca;
          when ADDx | ADDCx | SUBx | SUBCx | ANDx | ORx | XORx | CMPAx =>
--            nRD <= '0';
            b_ctrl <= B_LOAD;
            aout <= tpca + x;

          when LDAa =>
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            -- update zero flag
            if din = x"00"  then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if din(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            aout <= tpca;
          when LDAx =>
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            -- update zero flag
            if din = x"00"  then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if din(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            aout <= tpca + x;

          when STAa =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= acc;
            aout <= tpca;
          when STAx =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= acc;
            aout <= tpca + x;

          when BLDIVa =>
--            nRD <= '0';
            iv_ctrl <= IV_LDHI;
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;
          when BLDSPa =>
--            nRD <= '0';
            sp_ctrl <= SP_LDHI;
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;
          when BLDXa =>
--            nRD <= '0';
            x_ctrl <= X_LDHI;
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;
          when BSTSPa =>
            dout <= sp(15 downto 8);
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;
          when BSTXa =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= x(15 downto 8);
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;

          when LDAn | LDAxi | LDAix | STAn | STAxi | STAix |
               JMPn | JMPxi | JMPix | JSRn | JSRxi | JSRix =>
--            nRD <= '0';
            tpcb_ctrl <= TPCB_LDHI;
            tpca_ctrl <= TPCA_INC;
            aout <= tpca;

          when JSRa | JSRx =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc(7 downto 0);
            sp_ctrl <= SP_DEC;
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T8
      ----------------------------------------------------------------------
      when T8 =>
        case ir is
          when JIR =>
            sp_ctrl <= SP_DEC;
            irq_rst <= true;
            pc_ctrl <= PC_LOADIV;
            aout <= iv;
            rd_int_d <= '1';
          when RTI =>
            irq_rst <= true;
            iack_ctrl <= IACK_CLR;
            pc_ctrl <= PC_LOADA;
            aout <= tpca;
            rd_int_d <= '1';

          when ADDi | ADDa | ADDx =>
            acc_ctrl <= ACC_ADD;
            -- update carry flag
            if (acc+b) > 255 then sr_c_ctrl <= SR_C_SET;
                             else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if ((acc+b) and x"ff") = x"00"
                           then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if ((acc+b) and x"80") = x"80"
                           then sr_n_ctrl <= SR_N_SET;
                           else sr_n_ctrl <= SR_N_CLR;
            end if;
            -- update overflow flag
            if ((acc+b) > 255) xor (((acc and x"7F")+(b and x"7F")) > 127)
                           then sr_o_ctrl <= SR_O_SET;
                           else sr_o_ctrl <= SR_O_CLR;
            end if;
            rd_int_d <= '1';
          when ADDCi | ADDCa | ADDCx =>
            acc_ctrl <= ACC_ADDC;
            -- update carry flag
            if (acc+b+(sr and "00001")) > 255
                           then sr_c_ctrl <= SR_C_SET;
                           else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if ((acc+b+(sr and "00001")) and x"ff") = x"00"
                           then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if ((acc+b+(sr and "00001")) and x"80") = x"80"
                           then sr_n_ctrl <= SR_N_SET;
                           else sr_n_ctrl <= SR_N_CLR;
            end if;
            -- update overflow flag
            if ((acc+b+(sr and "00001")) > 255) xor
               (((acc and x"7F")+(b and x"7F")+(sr and "00001")) > 127)
                           then sr_o_ctrl <= SR_O_SET;
                           else sr_o_ctrl <= SR_O_CLR;
            end if;
            rd_int_d <= '1';
          when SUBi | SUBa | SUBx =>
            acc_ctrl <= ACC_SUB;
            -- update carry flag
            if (acc+not(b)+1) > 255
                           then sr_c_ctrl <= SR_C_SET;
                           else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if ((acc+not(b)+1) and x"ff") = x"00"
                           then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if ((acc+not(b)+1) and x"80") = x"80"
                           then sr_n_ctrl <= SR_N_SET;
                           else sr_n_ctrl <= SR_N_CLR;
            end if;
            -- update overflow flag
            if ((acc+not(b)+1) > 255) xor
               (((acc and x"7F")+not(b and x"7F")+1) > 127)
                           then sr_o_ctrl <= SR_O_SET;
                           else sr_o_ctrl <= SR_O_CLR;
            end if;
            rd_int_d <= '1';
          when SUBCi | SUBCa | SUBCx =>
            acc_ctrl <= ACC_SUBC;
            -- update carry flag
            if (acc+not(b)+(sr and "00001")) > 255
                           then sr_c_ctrl <= SR_C_SET;
                           else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if ((acc+not(b)+(sr and "00001")) and x"ff") = x"00"
                           then sr_z_ctrl <= SR_Z_SET;
                           else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if ((acc+not(b)+(sr and "00001")) and x"80") = x"80"
                           then sr_n_ctrl <= SR_N_SET;
                           else sr_n_ctrl <= SR_N_CLR;
            end if;
            -- update overflow flag
            if ((acc+not(b)+1) > 255) xor
               (((acc and x"7F")+not(b and x"7F")+(sr and "00001")) > 127)
                           then sr_o_ctrl <= SR_O_SET;
                           else sr_o_ctrl <= SR_O_CLR;
            end if;
            rd_int_d <= '1';
          when ANDi | ANDa | ANDx =>
            acc_ctrl <= ACC_AND;
            -- update zero flag
            if (acc and b) = x"00" then sr_z_ctrl <= SR_Z_SET;
                                   else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if (acc(7)='1') and (b(7)='1')
                                   then sr_n_ctrl <= SR_N_SET;
                                   else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when ORi | ORa | ORx =>
            acc_ctrl <= ACC_OR;
            -- update zero flag
            if (acc or b) = x"00"  then sr_z_ctrl <= SR_Z_SET;
                                   else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if (acc(7)='1') or (b(7)='1')
                                   then sr_n_ctrl <= SR_N_SET;
                                   else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when XORi | XORa | XORx =>
            acc_ctrl <= ACC_XOR;
            -- update zero flag
            if (acc xor b) = x"00" then sr_z_ctrl <= SR_Z_SET;
                                   else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if acc(7) /= b(7)      then sr_n_ctrl <= SR_N_SET;
                                   else sr_n_ctrl <= SR_N_CLR;
            end if;
            rd_int_d <= '1';
          when CMPAi | CMPAa | CMPAx =>
            -- update carry flag
            if acc > b then sr_c_ctrl <= SR_C_SET;
                       else sr_c_ctrl <= SR_C_CLR;
            end if;
            -- update zero flag
            if acc = b then sr_z_ctrl <= SR_Z_SET;
                       else sr_z_ctrl <= SR_Z_CLR;
            end if;
            rd_int_d <= '1';

          when LDAa | STAa | LDAx | STAx =>
            rd_int_d <= '1';

          when BSTSPa | BSTXa =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpca;

          when BLDIVa | BLDSPa | BLDXa|
               LDAn | LDAxi | LDAix | STAn | STAxi | STAix |
               JMPn | JMPxi | JMPix | JSRn | JSRxi | JSRix =>
            rd_int_d <= '1';
            aout <= tpca;

          when JSRa | JSRx =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T9
      ----------------------------------------------------------------------
      when T9 =>
        case ir is
          -- absolute addressing
          when BLDIVa =>
--            nRD <= '0';
            iv_ctrl <= IV_LDLO;
            aout <= tpca;
          when BLDSPa =>
--            nRD <= '0';
            sp_ctrl <= SP_LDLO;
            aout <= tpca;
          when BLDXa =>
--            nRD <= '0';
            x_ctrl <= X_LDLO;
            aout <= tpca;
          when BSTSPa =>
            dout <= sp(7 downto 0);
            aout <= tpca;
          when BSTXa =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= x(7 downto 0);
            aout <= tpca;

          when JSRa | JSRx =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc(15 downto 8);
            sp_ctrl <= SP_DEC;
            aout <= sp;

          when LDAn | LDAxi | LDAix | STAn | STAxi | STAix |
               JMPn | JMPxi | JMPix | JSRn | JSRxi | JSRix =>
--            nRD <= '0';
            tpcb_ctrl <= TPCB_LDLO;
            aout <= tpca;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T10
      ----------------------------------------------------------------------
      when T10 =>
        case ir is
          when BLDIVa | BLDSPa | BLDXa | BSTSPa | BSTXa =>
            rd_int_d <= '1';

          when LDAn | LDAxi =>
            rd_int_d <= '1';
            aout <= tpcb;
          when LDAix =>
            rd_int_d <= '1';
            aout <= tpcb + x;

          when STAn | STAxi =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpcb;
          when STAix =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= tpcb + x;

          when JMPn | JMPxi =>
            pc_ctrl <= PC_LOADB;
            aout <= tpcb;
            rd_int_d <= '1';
          when JMPix =>
            pc_ctrl <= PC_LOADBX;
            aout <= tpcb + x;
            rd_int_d <= '1';

          when JSRa =>
            pc_ctrl <= PC_LOADA;
            aout <= tpca;
            rd_int_d <= '1';
          when JSRx =>
            pc_ctrl <= PC_LOADAX;
            rd_int_d <= '1';
            aout <= tpca + x;
          when JSRn | JSRxi | JSRix =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T11
      ----------------------------------------------------------------------
      when T11 =>
        case ir is
          -- indirect addressing
          when LDAn | LDAxi =>
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            -- update zero flag
            if din = x"00"  then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if din(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            aout <= tpcb;
          when LDAix =>
--            nRD <= '0';
            acc_ctrl <= ACC_LOAD;
            -- update zero flag
            if din = x"00"  then sr_z_ctrl <= SR_Z_SET;
                            else sr_z_ctrl <= SR_Z_CLR;
            end if;
            -- update negative flag
            if din(7) = '1' then sr_n_ctrl <= SR_N_SET;
                            else sr_n_ctrl <= SR_N_CLR;
            end if;
            aout <= tpcb + x;

          when STAn | STAxi =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= acc;
            aout <= tpcb;
          when STAix =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= acc;
            aout <= tpcb + x;

          when JSRn | JSRxi | JSRix =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc(7 downto 0);
            sp_ctrl <= SP_DEC;
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T12
      ----------------------------------------------------------------------
      when T12 =>
        case ir is
          when LDAn | LDAxi | LDAix | STAn | STAxi | STAix =>
            rd_int_d <= '1';

          when JSRn | JSRxi | JSRix =>
            ddrive <= '1';
            wr_int_d <= '1';
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T13
      ----------------------------------------------------------------------
      when T13 =>
        case ir is
          when JSRn | JSRxi | JSRix =>
--            nWR <= '0';
            ddrive <= '1';
            dout <= pc(15 downto 8);
            sp_ctrl <= SP_DEC;
            aout <= sp;

          when others => NULL;
        end case;

      ----------------------------------------------------------------------
      -- CPU state T14
      ----------------------------------------------------------------------
      when T14 =>
        case ir is
          when JSRn | JSRxi =>
            pc_ctrl <= PC_LOADB;
            aout <= tpcb;
            rd_int_d <= '1';
          when JSRix =>
            pc_ctrl <= PC_LOADBX;
            aout <= tpcb + x;
            rd_int_d <= '1';

          when others => NULL;
        end case;

      when others => rd_int_d <= '1';
    end case;
  end process;  -- end output table
end behave;
